<?php
namespace Vitrium\Drm\HttpClient;

class HttpClient {
    private $curlHandle;
    private $options;

    // Base URL is optional. Endpoint is always appended when creating a request, and it's valid to pass a full URL as endpoint and leave baseUrl blank
    private $baseUrl;    

    private $request;
    private $response;

    protected $logger = null;

    // We parse the response headers using a CURL_HEADERFUNCTION callback function. The headers arrive as a series of strings, 
    // line by line, and end up parsed into this temporary store property before being converted into an array in the $response object
    private $responseHeaders;

    protected $helper;
    protected $config;

    /**
    * Public constructor
    *
    * @param array  $options  cUrl options
    * @param Logger $logger   Dependency injected instance of \Vitrium\Drm\Logger\Logger
    * @param Helper $helper   Dependency injected instance of \Vitrium\Drm\Helper\Data
    */
    public function __construct($options, \Vitrium\Drm\Logger\Logger $logger, \Vitrium\Drm\Helper\Data $helper) {
        if (!function_exists('curl_version'))
            throw new HttpClientException('cURL is not installed on this server', -1, new HttpRequest(), new HttpResponse());

        $this->config = parse_ini_file(dirname(__FILE__) . '/config.ini', true, INI_SCANNER_TYPED);

        $this->options = new CurlOptions($options);
        $this->baseUrl = rtrim($this->config['vitrium']['base_url'], '/');

        $this->logger = $logger;
        $this->helper = $helper;
    }    

    /**
    * Appends a query string to a given URL
    *
    * @param string $url        URL.
    * @param array  $parameters Query string parameters.
    */
    private function buildUrlQuery($url, $parameters = []) {
        if (!empty($parameters))
            $url .= '?' . http_build_query($parameters);

        return $url;
    }

    /**
    * Sets up the request method in cUrl options
    *
    * @param string $method Request method.
    */
    private function setupMethod($method) {
        if ('POST' == $method)
            curl_setopt($this->curlHandle, CURLOPT_POST, true);
        elseif (in_array($method, ['PUT', 'DELETE', 'OPTIONS']))
            curl_setopt($this->curlHandle, CURLOPT_CUSTOMREQUEST, $method);
    }

    /**
    * Creates and returns default request headers
    *
    * @param  mixed $body Request payload
    */
    protected function getRequestHeaders($body) {
        $headers = ['Accept' => 'application/json'];
        if (!empty($body)) {
            // Multipart requests will start with dashes and a boundary delimiter
            if (substr($body, 0, 5) == '-----') {
                $boundary = strstr($body, "\r\n", true);
                $boundary = substr($boundary, 2, strlen($boundary) - 2);
                $headers['Content-Type'] = "multipart/form-data; boundary=" . $boundary;
                $headers['Content-Length'] = strlen($body);
            }
            else {
                $headers['Content-Type'] = 'application/json;charset=utf-8';
            }
        }
        return $headers;
    }

    /**
    * Compose the request payload.
    */
    private function composeBody($data) {
        $this->logger->debug(json_encode($data));
        if (is_array($data) && !empty($data)) {
            // IMPORTANT: Start boundary with at least 3 slashes. getRequestHeaders determines if the content is multi-part by looking to see if body starts with dashes.
            $boundary = "----------" . md5(time());
            $eol = "\r\n";
            $payload = $boundary;

            foreach ($data as $dataPartName => $dataPart) {
                $payload .= $eol;
                // Set the Content-Disposition header
                $payload .= 'Content-Disposition: form-data; name="' . $dataPartName . '"';
                if (is_array($dataPart) && isset($dataPart['headers']) && is_array($dataPart['headers']) && isset($dataPart['headers']['Content-Disposition'])) {
                    $payload .= '; ' . $dataPart['headers']['Content-Disposition'];
                }
                $payload .= $eol;

                // Set any other headers for this content part
                if (is_array($dataPart) && isset($dataPart['headers']) && is_array($dataPart['headers'])) {
                    foreach ($dataPart['headers'] as $headerName => $headerValue) {
                        // Skip over Content-Disposition header since already been handled
                        if ($headerName == 'Content-Disposition')
                            continue;

                        $payload .= $headerName . ': ' . $headerValue . $eol;
                    }
                }

                $payload .= $eol;
                $payload .= (is_string($dataPart['data']) ? $dataPart['data'] : json_encode($dataPart['data'])) . $eol;
                $payload .= $boundary;
            }
            $payload .= '--' . $eol . $eol;
            return $payload;
        }
        else {
            return json_encode($data);
        }
    }

    /**
    * Sets up a cUrl request.
    *
    * @param string $endpoint   Request endpoint.
    * @param string $method     Request method.
    * @param array  $data       Request data.
    * @param array  $parameters Request parameters.
    *
    * @return Request
    */
    private function createRequest($endpoint, $method, $data = [], $parameters = []) {
        $body    = '';
        $url     = (isset($this->baseUrl) && !empty($this->baseUrl) ? rtrim($this->baseUrl, '/') . '/' : '') . $endpoint;

        // Setup method.
        $this->setupMethod($method);

        // Include post fields.
        if (!empty($data)) {
            $body = $this->composeBody($data);
            curl_setopt($this->curlHandle, CURLOPT_POSTFIELDS, $body);
        }

        $this->request = new HttpRequest(
            $this->buildUrlQuery($url, $parameters),
            $method,
            $parameters,
            $this->getRequestHeaders($body),
            $body
        );

        curl_setopt($this->curlHandle, CURLOPT_HTTPHEADER, $this->request->getRawHeaders());
        curl_setopt($this->curlHandle, CURLOPT_URL, $this->request->getUrl());

        $this->logger->debug('Request URL: ' . $this->request->getUrl());
        if (!empty($data) && !is_array($data)) {
            $this->logger->debug('Request payload: ' . $body);
        }

        return $this->getRequest();
    }

    /**
    * Get response headers.
    *
    * @return array
    */
    private function getResponseHeaders() {
        $headers = [];
        $lines   = explode("\n", $this->responseHeaders);
        $lines   = array_filter($lines, 'trim');

        foreach ($lines as $index => $line) {
            if (strpos($line, ': ') === false)
                 continue;

            list($key, $value) = explode(': ', $line);

            $headers[$key] = isset($headers[$key]) ? $headers[$key] . ', ' . trim($value) : trim($value);
        }

        return $headers;
    }

    /**
    * Set response headers.
    *
    * @return array
    */
    private function setResponseHeaders() {
        $this->responseHeaders = '';
        curl_setopt($this->curlHandle, CURLOPT_HEADERFUNCTION, function ($_, $headers) {
            $this->responseHeaders .= $headers;
            return strlen($headers);
        });
     }

    /**
    * Create response.
    *
    * @return HttpResponse
    */
    private function createResponse($body) {        
        $code    = curl_getinfo($this->curlHandle, CURLINFO_HTTP_CODE);
        $headers = $this->getResponseHeaders();

        // Create Response object
        $this->response = new HttpResponse($code, $headers, $body);

        return $this->getResponse();
    }

    /**
    * Initialize cUrl and set default cURL settings.
    */
    protected function initializeCurl() {
        $verifySsl       = $this->options->getVerifySsl();
        $timeout         = $this->options->getTimeout();
        $followRedirects = $this->options->getFollowRedirects();

        $this->curlHandle = curl_init();

        curl_setopt($this->curlHandle, CURLOPT_SSL_VERIFYPEER, $verifySsl);
        if (!$verifySsl) {
            curl_setopt($this->curlHandle, CURLOPT_SSL_VERIFYHOST, $verifySsl);
        }
        if ($followRedirects) {
            curl_setopt($this->curlHandle, CURLOPT_FOLLOWLOCATION, true);
        }
        curl_setopt($this->curlHandle, CURLOPT_CONNECTTIMEOUT, $timeout);
        curl_setopt($this->curlHandle, CURLOPT_TIMEOUT, $timeout);
        curl_setopt($this->curlHandle, CURLOPT_RETURNTRANSFER, true);
    }

    /**
    * Look for errors in the request.
    *
    * @param array $parsedResponse Parsed body response.
    */
    private function lookForErrors($parsedResponse) {
        // Any non-200/201/202 response code indicates an error.
        if (!in_array($this->response->getCode(), ['200', '201', '202'])) {
            $errors = isset($parsedResponse->errors) ? $parsedResponse->errors : $parsedResponse;

            if (is_array($errors) && is_object($errors[0])) {
                 $errorMessage = $errors[0]->message;
                 $errorCode    = $errors[0]->code;
            } else if (is_object($errors)) {
                 $errorMessage = property_exists($errors, 'message') ? $errors->message : json_encode($parsedResponse);
                 $errorCode    = property_exists($errors, 'code') ? $errors->code : 'Unknown code';
            }
            else {
                 $errorMessage = $parsedResponse;
                 $errorCode    = 'Unknown error code';
            }

            throw new HttpClientException(
                 sprintf('Error: %s [%s] HTTP Status [%s]', $errorMessage, $errorCode, $this->response->getCode()),
                 $this->response->getCode(),
                 $this->request,
                 $this->response
            );
        }
    }

    /**
    * Process response.
    *
    * @return array
    */
    private function processResponse() {
        $body = $this->response->getBody();

        if (0 === strpos(bin2hex(substr($body, 0, 4)), 'efbbbf')) {
            $body = substr($body, 3);
        }

        $parsedResponse = json_decode($body);

        // Test for valid JSON.
        if (JSON_ERROR_NONE !== json_last_error()) {
            $message = function_exists('json_last_error_msg') ? json_last_error_msg() : 'Invalid JSON returned';
            throw new HttpClientException($message, $this->response->getCode(), $this->request, $this->response);
        }

        $this->logger->debug('Response to URL ' . $this->request->getUrl() . ': [' . $this->response->getCode() . '] ' . $body);
        $this->lookForErrors($parsedResponse);

        return $parsedResponse;
     }

    /**
    * Main method for dipatching requests
    *
    * @param string $endpoint   Request endpoint.
    * @param string $method     Request method.
    * @param array  $data       Request data.
    * @param array  $parameters Request parameters.
    *
    * @return array
    */
    public function request($endpoint, $method, $data = [], $parameters = []) {
        // Initialize cUrl.
        $this->initializeCurl();

        // Set request args.
        $request = $this->createRequest($endpoint, $method, $data, $parameters);

        $this->setResponseHeaders();

        // Get response.        
        $response = $this->createResponse(curl_exec($this->curlHandle));

        // Check for cURL errors.
        if (curl_errno($this->curlHandle)) {
            throw new HttpClientException('cURL Error: ' . curl_error($this->curlHandle), 0, $request, $response);
        }

        curl_close($this->curlHandle);

        $response = $this->processResponse();

        return $response;
    }

    public function getRequest() {
        return $this->request;
    }
    public function getResponse() {
        return $this->response;
    }

    /**
    * Shortcut functions that issue requests by type.
    * @param string $endpoint   API endpoint (required for all shortcut functions).
    * @param array  $data       Request data for POST and PUT; must be supplied but can be null or empty.
    * @param array  $parameters Optional parameters for GET and DELETE.
    *
    * @return stdClass JSON-decoded response body
    */
    public function post($endpoint, $data) {
        return $this->request($endpoint, 'POST', $data);
    }
    public function put($endpoint, $data) {
        return $this->request($endpoint, 'PUT', $data);
    }
    public function get($endpoint, $parameters = []) {
        return $this->request($endpoint, 'GET', [], $parameters);
    }
    public function delete($endpoint, $parameters = []) {
        return $this->request($endpoint, 'DELETE', [], $parameters);
    }
    public function options($endpoint) {
        return $this->request($endpoint, 'OPTIONS', [], []);
    }
}
